/*
* Copyright (c) 2017-2024 Nils Kohl, Daniel Bauer, Fabian Böhm.
*
* This file is part of HyTeG
* (see https://i10git.cs.fau.de/hyteg/hyteg).
*
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

/*
* The entire file was generated with the HyTeG Operator Generator.
*
* Avoid modifying this file. If buggy, consider fixing the generator itself.
*/

#pragma once

#include "core/DataTypes.h"

#include "hyteg/LikwidWrapper.hpp"
#include "hyteg/communication/Syncing.hpp"
#include "hyteg/edgedofspace/EdgeDoFMacroCell.hpp"
#include "hyteg/geometry/AnnulusMap.hpp"
#include "hyteg/operators/Operator.hpp"
#include "hyteg/p2functionspace/P2Function.hpp"
#include "hyteg/p2functionspace/P2VectorFunction.hpp"
#include "hyteg/primitivestorage/PrimitiveStorage.hpp"
#include "hyteg/solvers/Smoothables.hpp"
#include "hyteg/sparseassembly/SparseMatrixProxy.hpp"

#define FUNC_PREFIX

namespace hyteg {

namespace operatorgeneration {

/// "Epsilon" operator.
///
/// Geometry map:    AnnulusMap
///
/// Weak formulation
///
///     u: trial function (vectorial space: TensorialVectorSpace(Lagrange, degree: 2))
///     v: test function  (vectorial space: TensorialVectorSpace(Lagrange, degree: 2))
///     μ: coefficient    (scalar space:    Lagrange, degree: 2)
///
///     ∫ 2 μ ε(u) : ε(v)
///
/// where
///
///     ε(w) := (1/2) (∇w + (∇w)ᵀ)

class P2VectorElementwiseEpsilonRotationAnnulusMap : public Operator< P2VectorFunction< real_t >, P2VectorFunction< real_t > >,
                                                     public OperatorWithInverseDiagonal< P2VectorFunction< real_t > >
{
 public:
   P2VectorElementwiseEpsilonRotationAnnulusMap( const std::shared_ptr< PrimitiveStorage >& storage,
                                                 size_t                                     minLevel,
                                                 size_t                                     maxLevel,
                                                 const P2Function< real_t >&                _mu,
                                                 const P2Function< real_t >&                _nx,
                                                 const P2Function< real_t >&                _ny );

   void apply( const P2VectorFunction< real_t >& src,
               const P2VectorFunction< real_t >& dst,
               uint_t                            level,
               DoFType                           flag,
               UpdateType                        updateType = Replace ) const;

   void toMatrix( const std::shared_ptr< SparseMatrixProxy >& mat,
                  const P2VectorFunction< idx_t >&            src,
                  const P2VectorFunction< idx_t >&            dst,
                  uint_t                                      level,
                  DoFType                                     flag ) const;

   void computeInverseDiagonalOperatorValues();

   std::shared_ptr< P2VectorFunction< real_t > > getInverseDiagonalValues() const;

 protected:
 private:
   /// Kernel type: apply
   /// - quadrature rule: Dunavant 3 | points: 4, degree: 3
   /// - operations per element:
   ///   adds    muls    divs    pows    abs    assignments    function_calls    unknown_ops
   /// ------  ------  ------  ------  -----  -------------  ----------------  -------------
   ///   1410    2294      20      18      0              0                 0              0
   void apply_macro_2D( real_t* RESTRICT _data_dst_edge_0,
                        real_t* RESTRICT _data_dst_edge_1,
                        real_t* RESTRICT _data_dst_vertex_0,
                        real_t* RESTRICT _data_dst_vertex_1,
                        real_t* RESTRICT _data_muEdge,
                        real_t* RESTRICT _data_muVertex,
                        real_t* RESTRICT _data_nxEdge,
                        real_t* RESTRICT _data_nxVertex,
                        real_t* RESTRICT _data_nyEdge,
                        real_t* RESTRICT _data_nyVertex,
                        real_t* RESTRICT _data_src_edge_0,
                        real_t* RESTRICT _data_src_edge_1,
                        real_t* RESTRICT _data_src_vertex_0,
                        real_t* RESTRICT _data_src_vertex_1,
                        real_t           macro_vertex_coord_id_0comp0,
                        real_t           macro_vertex_coord_id_0comp1,
                        real_t           macro_vertex_coord_id_1comp0,
                        real_t           macro_vertex_coord_id_1comp1,
                        real_t           macro_vertex_coord_id_2comp0,
                        real_t           macro_vertex_coord_id_2comp1,
                        int64_t          micro_edges_per_macro_edge,
                        real_t           micro_edges_per_macro_edge_float,
                        real_t           radRayVertex,
                        real_t           radRefVertex,
                        real_t           rayVertex_0,
                        real_t           rayVertex_1,
                        real_t           refVertex_0,
                        real_t           refVertex_1,
                        real_t           thrVertex_0,
                        real_t           thrVertex_1 ) const;
   /// Kernel type: toMatrix
   /// - quadrature rule: Dunavant 3 | points: 4, degree: 3
   /// - operations per element:
   ///   adds    muls    divs    pows    abs    assignments    function_calls    unknown_ops
   /// ------  ------  ------  ------  -----  -------------  ----------------  -------------
   ///   1266    2150      20      18      0              0                 0              3
   void toMatrix_macro_2D( idx_t* RESTRICT                      _data_dst_edge_0,
                           idx_t* RESTRICT                      _data_dst_edge_1,
                           idx_t* RESTRICT                      _data_dst_vertex_0,
                           idx_t* RESTRICT                      _data_dst_vertex_1,
                           real_t* RESTRICT                     _data_muEdge,
                           real_t* RESTRICT                     _data_muVertex,
                           real_t* RESTRICT                     _data_nxEdge,
                           real_t* RESTRICT                     _data_nxVertex,
                           real_t* RESTRICT                     _data_nyEdge,
                           real_t* RESTRICT                     _data_nyVertex,
                           idx_t* RESTRICT                      _data_src_edge_0,
                           idx_t* RESTRICT                      _data_src_edge_1,
                           idx_t* RESTRICT                      _data_src_vertex_0,
                           idx_t* RESTRICT                      _data_src_vertex_1,
                           real_t                               macro_vertex_coord_id_0comp0,
                           real_t                               macro_vertex_coord_id_0comp1,
                           real_t                               macro_vertex_coord_id_1comp0,
                           real_t                               macro_vertex_coord_id_1comp1,
                           real_t                               macro_vertex_coord_id_2comp0,
                           real_t                               macro_vertex_coord_id_2comp1,
                           std::shared_ptr< SparseMatrixProxy > mat,
                           int64_t                              micro_edges_per_macro_edge,
                           real_t                               micro_edges_per_macro_edge_float,
                           real_t                               radRayVertex,
                           real_t                               radRefVertex,
                           real_t                               rayVertex_0,
                           real_t                               rayVertex_1,
                           real_t                               refVertex_0,
                           real_t                               refVertex_1,
                           real_t                               thrVertex_0,
                           real_t                               thrVertex_1 ) const;
   /// Kernel type: computeInverseDiagonalOperatorValues
   /// - quadrature rule: Dunavant 3 | points: 4, degree: 3
   /// - operations per element:
   ///   adds    muls    divs    pows    abs    assignments    function_calls    unknown_ops
   /// ------  ------  ------  ------  -----  -------------  ----------------  -------------
   ///    666    1086      20      18      0              0                 0              0
   void computeInverseDiagonalOperatorValues_macro_2D( real_t* RESTRICT _data_invDiag__edge_0,
                                                       real_t* RESTRICT _data_invDiag__edge_1,
                                                       real_t* RESTRICT _data_invDiag__vertex_0,
                                                       real_t* RESTRICT _data_invDiag__vertex_1,
                                                       real_t* RESTRICT _data_muEdge,
                                                       real_t* RESTRICT _data_muVertex,
                                                       real_t* RESTRICT _data_nxEdge,
                                                       real_t* RESTRICT _data_nxVertex,
                                                       real_t* RESTRICT _data_nyEdge,
                                                       real_t* RESTRICT _data_nyVertex,
                                                       real_t           macro_vertex_coord_id_0comp0,
                                                       real_t           macro_vertex_coord_id_0comp1,
                                                       real_t           macro_vertex_coord_id_1comp0,
                                                       real_t           macro_vertex_coord_id_1comp1,
                                                       real_t           macro_vertex_coord_id_2comp0,
                                                       real_t           macro_vertex_coord_id_2comp1,
                                                       int64_t          micro_edges_per_macro_edge,
                                                       real_t           micro_edges_per_macro_edge_float,
                                                       real_t           radRayVertex,
                                                       real_t           radRefVertex,
                                                       real_t           rayVertex_0,
                                                       real_t           rayVertex_1,
                                                       real_t           refVertex_0,
                                                       real_t           refVertex_1,
                                                       real_t           thrVertex_0,
                                                       real_t           thrVertex_1 ) const;

   std::shared_ptr< P2VectorFunction< real_t > > invDiag_;
   P2Function< real_t >                          mu;
   P2Function< real_t >                          nx;
   P2Function< real_t >                          ny;
};

} // namespace operatorgeneration

} // namespace hyteg
